// ******** Graphics

DWORD screen[320*240];

void DrawScreen(DWORD* buffer) // scale 320x240 -> 640x480
{
	for(int y=0;y<480;y++)
	{
		DWORD* sp=screen+y/2*320;
		for(int x=0;x<640;x+=2)
		{
			*buffer++=*sp;
			*buffer++=*sp++;
		}
	}
}

void ClearScreen()
{
	for(int i=0;i<320*240;i++)
		screen[i]=0x000000;
}

char spr_player[8*8+1]=	"...xx..."
						"..x..x.."
						"..x..x.."
						"..xxxx.."
						"...xx..."
						".xxxxxx."
						".xxxxxx."
						"..x..x..";

char spr_growth[3*8*8+1]=	"........"
							"........"
							"..x..x.."
							"...xx..."
							"...xx..."
							"..x..x.."
							"........"
							"........"

							"........"
							"........"
							"..x..x.."
							".xxxxxx."
							"...xx..."
							"..xxxx.."
							"........"
							"........"

							"........"
							"........"
							".xxxxxx."
							".x....x."
							"..x..x.."
							".x.xx.x."
							"...xx..."
							"........";

char spr_loot[8*8+1]=	"........"
						".xxxxxx."
						".x....x."
						".x.xx.x."
						".x.xx.x."
						".x....x."
						".xxxxxx."
						"........";

char spr_font[10*8*8+1]="xxxxxx.."
						"x....x.."
						"x....x.."
						"x....x.."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"..xx...."
						"...x...."
						"...x...."
						"...x...."
						".xxxxx.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						".....x.."
						"xxxxxx.."
						"x......."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						".....x.."
						"...xxx.."
						".....x.."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"x....x.."
						"x....x.."
						"xxxxxx.."
						".....x.."
						".....x.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						"x......."
						"xxxxxx.."
						".....x.."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						"x......."
						"xxxxxx.."
						"x....x.."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						".....x.."
						".....x.."
						".....x.."
						".....x.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						"x....x.."
						"xxxxxx.."
						"x....x.."
						"xxxxxx.."
						"........"
						"........"
						"........"

						"xxxxxx.."
						"x....x.."
						"xxxxxx.."
						".....x.."
						"xxxxxx.."
						"........"
						"........"
						"........";


void DrawSprite(char* sprite, DWORD color, int sx, int sy, int dir)
{
	int rv[4*6]={0,0, 1,0, 0,1,
				 7,0, 0,1,-1,0,
				 7,7,-1,0, 0,-1,
				 0,7, 0,-1, 1,0};
	int x=sx+rv[dir*6+0], y=sy+rv[dir*6+1];
	int xdx= rv[dir*6+2], xdy= rv[dir*6+3];
	int ydx= rv[dir*6+4], ydy= rv[dir*6+5];
	for(int ty=0;ty<8;ty++)
	{
		for(int tx=0;tx<8;tx++)
		{
			if(sprite[ty*8+tx]!='.' && x>=0 && x<320 && y>=0 && y<240)
				screen[y*320+x]=color;
			x+=xdx; y+=xdy;
		}
		x-=xdx*8; y-=xdy*8;
		x+=ydx; y+=ydy;
	}
}

void DrawBar(int x1, int y1, int w, int h, DWORD color)
{
	for(int y=y1;y<y1+h;y++)
	{
		DWORD* dp=screen+y*320+x1;
		for(int x=x1;x<x1+w;x++)
			*dp++=color;
	}
}

void DrawNumber(int value, int x, int y, DWORD color)
{
	for(int i=0;i<6;i++)
	{
		int digit=value%10;
		DrawSprite(spr_font+digit*64, color, x-i*7-8, y, 0);
		value/=10;
	}
}


// ******** Audio

struct Channel
{
	float wave_y[2];
	float wave_dy[2];
	float wdy;
	int wpos;
	bool noise;
	float vol;
	float duty;
	float ffrq;
	float fdmp;
	float period;
	int iperiod;
	int iperiodh;
	int tick;
	float rp;
	float p;
	float dp;
	bool decay;
	float decspd;
	bool attack;
	float attspd;
};

Channel chan[8];

float master_vol=0.15f;

float nper[120];

int song_delay;
int song_speed;
int song_pos;
int song_bigpos;
int song_scale;
int song_key;
int song_patlen;
int song_pattern[32];
int song_beat[32];
int song_scales[3][6];
int song_scalen[3];
int song_scalemap[4];
int song_keymap[4];
bool song_octave;

bool beat_trigger=false;

float vbuffer[320];
int vbi=0;

void NoteOn(int ci, int note, float vol)
{
	Channel& ch=chan[ci];
	if(ch.vol==0.0f)
		ch.vol=0.01f;
	ch.period=nper[note];
	ch.decay=false;
	ch.attack=true;
}

void NoteOff(int ci)
{
	Channel& ch=chan[ci];
	ch.decay=true;
}

void ChangeChannel(int i, float cutoff, float reso, float duty)
{
	Channel& ch=chan[i];
	ch.duty=duty;
	ch.ffrq=0.00002f*(1.0f+cutoff*cutoff*1000);
	ch.fdmp=(0.001f-0.0008f*cutoff)+(1.0f-sqrt(reso))*0.2f;
}

void RenderChannel(Channel& ch, float* buffer, int size)
{
	for(int i=0;i<size*4;i++)
	{
		// wave shape
		ch.tick++;
		if(ch.tick==ch.iperiodh)
		{
			ch.wpos=1;
			ch.rp=ch.wave_y[ch.wpos];
			ch.wdy=ch.wave_dy[ch.wpos]/(ch.iperiod-ch.iperiodh);
		}
		if(ch.tick>=ch.iperiod)
		{
			ch.wpos=0;
			ch.tick=0;
			ch.iperiod=(int)ch.period;
			ch.iperiodh=(int)(ch.period*(0.5+ch.duty/2));
			ch.rp=ch.wave_y[ch.wpos];
			ch.wdy=ch.wave_dy[ch.wpos]/ch.iperiodh;
		}
		// target signal level
		ch.rp+=ch.wdy;
		if(ch.noise)
			ch.rp=(float)(rnd(100)-50)/150;
		// resonant lowpass filter
		ch.dp+=(ch.rp-ch.p)*ch.ffrq;
		ch.dp-=ch.dp*ch.fdmp;
		ch.p+=ch.dp;
		if(ch.tick%8==0)
		{
			if(&ch==&chan[3])
				ch.period*=1.001f;
			if(ch.decay)
			{
				ch.vol-=ch.decspd;
				if(ch.vol<0.0001f)
				{
					ch.vol=0.0f;
					ch.decay=false;
				}
			}
			if(ch.attack)
			{
				ch.vol+=ch.attspd;
				if(ch.vol>0.2f)
				{
					ch.vol=0.2f;
					ch.attack=false;
				}
			}
		}
		// 4x supersampling
		*buffer+=ch.p*ch.vol;
		if(i%4==3)
			buffer++;
	}
}

void NewBeat()
{
	for(int i=0;i<32;i++)
	{
		song_beat[i]=0;
		if(i%4==0)
			song_beat[i]=1;
		else if(i%2==0 && rnd(5)!=0)
			song_beat[i]|=2;

		if(i%4==2 && rnd(5)==0)
			song_beat[i]|=1;

		if(i%4==0 && rnd(7)==0)
			song_beat[i]|=2;
	}
}

void NewPattern()
{
	int curnote=rnd(6)-3;
	int pnote=curnote;
	for(int i=0;i<32;i++)
	{
		song_pattern[i]=-1;
		bool note_on=false;
		if(i%4==0 && rnd(100)<90)
			note_on=true;
		else if(i%2==0 && rnd(100)<75)
			note_on=true;
		else if(rnd(100)<15)
			note_on=true;
		if(i%4==0)
		{
			if(rnd(2)==0)
				curnote=pnote;
			pnote=curnote;
		}
		if(note_on)
		{
			song_pattern[i]=curnote;
			if(rnd(1))
				curnote++;
			else
				curnote--;
		}
	}
	song_patlen=(rnd(3)+1)*8;
}

void SongTick()
{
	NoteOff(0);
	int rsp=song_pos%song_patlen;
	if(song_pattern[rsp]!=-1)
	{
		int sp=song_pattern[rsp];
		int sl=song_scalen[song_scale];
		NoteOn(0, song_scales[song_scale][sp%sl]+sp/sl*12+24+song_key, 0.2f);
	}

	NoteOff(3);
	if(song_beat[song_pos]&1)
		NoteOn(3, 50, 0.2f);
	if(song_beat[song_pos]&2)
		NoteOn(5, 50, 0.015f);
	NoteOff(5);
	
	if(song_pos%4==0)
	{
		NoteOn(1, 12+song_key, 0.2f);
		beat_trigger=true;
	}
	if(song_pos%4==2)
	{
		if(song_octave)
			NoteOn(1, 24+song_key, 0.2f);
		else
			NoteOff(1);
	}

	if(song_pos%32==0 && (rnd(3)==0 || song_bigpos==0))
		NoteOn(2, 36+song_key+song_scales[song_scale][rnd(2)], 0.2f);
	if(song_pos%32==3)
		NoteOff(2);

	song_pos++;
	if(song_pos>=32)
	{
		song_pos=0;
		song_bigpos++;
		if(rnd(2)==0)
			NewPattern();
		if(song_bigpos==4)
		{
			NewPattern();
			song_bigpos=0;
			if(rnd(2)==0)
				song_octave=!song_octave;
			if(rnd(4)==0)
			{
				int d=rnd(2)-1;
				for(int i=0;i<4;i++)
				{
					song_scalemap[i]=rnd(2);
					song_keymap[i]+=d;
					if(song_keymap[i]<0) song_keymap[i]=rnd(11);
				}
			}
			if(rnd(1))
				NewBeat();
		}
		song_scale=song_scalemap[song_bigpos];
		song_key=song_keymap[song_bigpos];
	}
}

void AudioStream(float* buffer, int size)
{
	srand(timeGetTime());
	float* bp=buffer;
	for(int i=0;i<size;i++)
		*bp++=0.0f;
	for(int bi=0;bi<size/128;bi++)
	{
		song_delay++;
		if(song_delay==song_speed)
		{
			song_delay=0;
			SongTick();
		}
		// mix all active audio channels into one
		for(int ci=0;ci<8;ci++)
			if(chan[ci].vol>0.0f)
				RenderChannel(chan[ci], buffer+bi*128, 128);
	}
	// scale to master volume
	bp=buffer;
	for(int i=0;i<size;i++)
	{
		vbuffer[vbi/3]=*bp;
		vbi=(vbi+1)%(320*3);
		*bp++*=master_vol;
	}
}



// ******** Logic

int ddx[4]={0,1,0,-1};
int ddy[4]={-1,0,1,0};

int count=0;

bool jukebox=false;

int plr_x;
int plr_y;
int plr_dir;
int plr_integrity;
int plr_boost;
int plr_score;
int plr_dead;

int gmap[40*30];
float ft, xsc, ysc;

float gamespeed;

struct Loot
{
	bool is;
	int x;
	int y;
};

Loot loot[100];

float ps=0.0f;
int rotbase=0;

void ResetGame()
{
	plr_x=160-4;
	plr_y=150;
	plr_dir=0;
	plr_integrity=120;
	plr_boost=0;
	plr_score=0;
	plr_dead=0;

	ft=(float)rnd(1000)*0.1f;
	xsc=0.0f;
	ysc=0.0f;
	
	for(int i=0;i<40*30;i++)
		gmap[i]=0;
	for(int i=0;i<100;i++)
		loot[i].is=false;
	
	gamespeed=0.0003f;
	count=0;

	for(int i=0;i<8;i++)
	{
		chan[i].vol=0.0f;
		chan[i].tick=0;
		chan[i].rp=0.0f;
		chan[i].p=0.0f;
		chan[i].dp=0.0f;
		chan[i].attspd=0.0004f;
		chan[i].decspd=0.00003f;
		chan[i].attack=false;
		chan[i].decay=false;
		ChangeChannel(i, 1.0f, 0.0f, 0.0f);
	}

	chan[0].wave_y[0]=1.0f;
	chan[0].wave_dy[0]=-(float)rnd(10)/10;
	chan[0].wave_y[1]=-1.0f;
	chan[0].wave_dy[1]=(float)rnd(10)/10;;
	
	chan[1].wave_y[0]=1.0f;
	chan[1].wave_dy[0]=-1.0f;
	chan[1].wave_y[1]=0.0f;
	chan[1].wave_dy[1]=-1.0f;
	chan[1].decspd=0.001f;
	ChangeChannel(1, 0.5f, 0.6f, 0.0);

	chan[2].wave_y[0]=1.0f;
	chan[2].wave_dy[0]=-1.0f;
	chan[2].wave_y[1]=0.0f;
	chan[2].wave_dy[1]=0.0f;
	chan[2].attspd=0.00002f;
	chan[2].decspd=0.000005f;
	ChangeChannel(2, 0.7f, 0.2f, 0.0);

	chan[3].wave_y[0]=1.0f;
	chan[3].wave_dy[0]=-2.0f;
	chan[3].wave_y[1]=-1.0f;
	chan[3].wave_dy[1]=2.0f;
	chan[3].attspd=0.002f;
	chan[3].decspd=0.0005f;
	ChangeChannel(3, 0.3f, 0.6f, 0.0);

	chan[4].wave_y[0]=0.75f;
	chan[4].wave_dy[0]=0.0f;
	chan[4].wave_y[1]=-0.75f;
	chan[4].wave_dy[1]=0.0f;
	chan[4].attspd=0.02f;
	chan[4].decspd=0.00005f;
	ChangeChannel(4, 1.0f, 0.4f, 0.0);

	chan[5].noise=true;
	chan[5].attspd=0.02f;
	chan[5].decspd=0.00005f;
	ChangeChannel(5, 1.0f, 0.6f, 0.0);

	chan[6].wave_y[0]=1.0f;
	chan[6].wave_dy[0]=-2.0f;
	chan[6].wave_y[1]=-1.0f;
	chan[6].wave_dy[1]=0.0f;
	chan[6].decspd=0.001f;
	ChangeChannel(6, 0.2f, 0.9f, 0.0);

	ps=(float)rnd(1000)/100;

	double freq=55.0/2;
	for(int i=0;i<120;i++)
	{
		nper[i]=44100.0f*4/(float)freq;
		freq*=1.05946309436;
	}
	
	song_scalen[0]=3;
	song_scales[0][0]=0;
	song_scales[0][1]=4;
	song_scales[0][2]=7;

	song_scalen[1]=3;
	song_scales[1][0]=0;
	song_scales[1][1]=3;
	song_scales[1][2]=7;

	song_scalen[2]=3;
	song_scales[2][0]=0;
	song_scales[2][1]=5;
	song_scales[2][2]=7;
	
	NewPattern();
	for(int i=0;i<4;i++)
	{
		song_scalemap[i]=rnd(2);
		song_keymap[i]=rnd(11);
	}
	song_delay=0;
	song_speed=37+rnd(10);
	song_pos=0;
	song_bigpos=0;
	song_scale=song_scalemap[0];
	song_key=song_keymap[0];
	if(rnd(1))
		song_octave=true;
	else
		song_octave=false;
	NewBeat();
	
	master_vol=0.2f;
}

bool GameTick() // to be called at 60 fps
{
	if(KeyDown(KEY_START) && !jukebox)
	{
		jukebox=true;
		count=0;
	}

	if(jukebox)
	{
		gamespeed=0.01f;
		if(beat_trigger)
		{
			rotbase++;
			beat_trigger=false;
		}
		Sleep(15);
	}

	if(jukebox && KeyDown(KEY_START) && count>30)
		return false;

	ClearScreen();

	// sweeping audio filter
	ps+=0.01f;
	ChangeChannel(0, 0.5f+sin(ps)*0.3f, 0.9f, 0.6f+cos(ps*1.3f)*0.3f);

	if((plr_dead>60 || (jukebox && count>30)) && KeyDown(KEY_FIRE))
	{
		count=0;
		ResetGame();
	}

	int plr_mi=(plr_y+4)/8*40+(plr_x+4)/8;
	count++;
	if(plr_dead==0 && !jukebox)
	{
		// player movement
		int spd=1;
		if(gmap[plr_mi]!=0 && (count&1)) spd=0;
		if(gmap[plr_mi]==2)
		{
			plr_integrity--;
			chan[6].vol=0.25f;
			chan[6].period=5000.0f+rnd(2000);
			ChangeChannel(6, 0.35f, 0.9f, 0.0);
		}
		else
			chan[6].vol=0.0f;
		if(gmap[plr_mi]==0)
			plr_integrity++;
		int ndir=-1;
		if(KeyDown(KEY_UP)) ndir=0;
		if(KeyDown(KEY_RIGHT)) ndir=1;
		if(KeyDown(KEY_DOWN)) ndir=2;
		if(KeyDown(KEY_LEFT)) ndir=3;
		if(KeyDown(KEY_FIRE) && plr_boost>0)
		{
			if(ndir!=-1)
				plr_boost--;
			spd=2;
		}
		if(ndir!=-1)
		{
			plr_x+=ddx[plr_dir]*spd;
			plr_y+=ddy[plr_dir]*spd;
			plr_dir=ndir;
		}
		if(plr_x<0) plr_x=0;
		if(plr_x>312) plr_x=312;
		if(plr_y<0) plr_y=0;
		if(plr_y>232) plr_y=232;
		if(plr_integrity>120) plr_integrity=120;
		if(plr_integrity<0)
		{
			plr_integrity=0;
			plr_dead=1;
			chan[6].vol=0.15f;
			chan[6].period=150.0f;
			ChangeChannel(6, 0.8f, 0.7f, 0.0);
		}
	}
	else if(!jukebox)
	{
		plr_dead++;
		chan[6].period=150.0f+rnd(100);
		master_vol-=0.001f;
		if(master_vol<0.0f)
			master_vol=0.0f;
	}

	// map
	ft+=gamespeed;
	xsc=cos(ft/13)*7.0f;
	ysc=sin(ft/25)*7.0f;
	DWORD bgpal[3]={0x80A040, 0x4080A0, 0xA04080};
	for(int y=0;y<30;y++)
	{
		float f=cos((float)y*0.13f)+sin(ft)*2.5f;
		for(int x=0;x<40;x++)
		{
			float df=f+cos((float)x*0.168f+xsc)*0.5f+sin((float)y*0.218f+ysc)*0.5f+cos(f)*0.47f+x*0.03f;
			int i=abs((int)(sin(df*0.1127f)*1600.0f))%400;
			if(i>=300) i-=200;
			int linc=(int)(gamespeed*15000);
			if(linc>70) linc=70;
			i+=linc;
			i/=100;
			if(i>2) i=2;
			gmap[y*40+x]=i;
			DrawSprite(&spr_growth[i*64], bgpal[i], x*8, y*8, (y+x/8+rotbase)%4);
		}
	}

	// loot
	if(rnd(15)==0 || jukebox)
		for(int i=0;i<100;i++)
			if(!loot[i].is)
			{
				loot[i].is=true;
				loot[i].x=rnd(39);
				loot[i].y=rnd(29);
				if(gmap[loot[i].y*40+loot[i].x]!=2)
					loot[i].is=false;
				else
					break;
			}
	DWORD lcolor=0xFFFFFF;
	if(count%16<8) lcolor=0xFFC080;
	for(int i=0;i<100;i++)
		if(loot[i].is)
		{
			DrawSprite(spr_loot, lcolor, loot[i].x*8, loot[i].y*8, 0);
			int lmi=loot[i].y*40+loot[i].x;
			if(gmap[lmi]!=2)
				loot[i].is=false;
			if(plr_dead==0 && plr_x+8>loot[i].x*8 && plr_x<loot[i].x*8+8 && plr_y+8>loot[i].y*8 && plr_y<loot[i].y*8+8)
			{
				loot[i].is=false;
				plr_boost+=20;
				plr_score+=16;
				gamespeed*=1.02f;
				if(!jukebox)
				{
					NoteOn(4, 80, 0.2f);
					NoteOff(4);
				}
			}
		}

	if(plr_boost>60) plr_boost=60;
	while(count==1 && gmap[plr_mi]==2)
	{
		plr_x=(rnd(33)+3)*8;
		plr_y=(rnd(23)+3)*8;
		plr_mi=(plr_y+4)/8*40+(plr_x+4)/8;
	}

	if(!jukebox)
	{
		if(plr_dead==0)
		{	
			// player sprite
			DrawSprite(spr_player, 0xC0FFFF, plr_x, plr_y, plr_dir);
		}
		
		// stats
		DrawBar(97, 1, 126, 14, 0x000000);
		DrawBar(98, 2, 124, 12, bgpal[1]);
		DrawBar(99, 3, 122, 4, 0x000000);
		DrawBar(100, 4, plr_integrity, 2, 0xFFA040);
		DrawBar(99, 8, 62, 5, 0x000000);
		DrawBar(100, 9, plr_boost, 3, 0xFFFFFF);
		DrawNumber(plr_score, 214, 8, 0xFFFFFF);
	}
	else
	{
		DrawBar(0, 120-40, 320, 80, 0x000000);
		DrawBar(0, 120-39, 320, 1, bgpal[1]);
		DrawBar(0, 120+38, 320, 1, bgpal[1]);
		int p=(int)(vbuffer[(i+vbi/3)%320]*10);
		for(int i=0;i<320;i++)
		{
			int y2=p;
			p=(int)(vbuffer[(i+vbi/3)%320]*10);
			if(p<-32) p=-32;
			if(p>31) p=31;
			int y1=p;
			if(y2<y1)
			{
				int t=y2;
				y2=y1;
				y1=t;
			}
			if(y2==y1)
				y2++;
			DrawBar(i, 120+y1, 1, y2-y1, 0xFFFFFF);
		}
	}

	Sleep(10);

	return true;
}

